#!/usr/bin/env python
# -*- coding: utf-8 -*- 
#
# Copyright(c) 2017 NEC Corporation.
#
# Tool to create various VH information files from the host information
# file of Excel format.
#
# % create-hostlist.py --type { ansible | cobbler | power_bmc | power_wol } 
#                 	  [--infile filename][ --outfile  filename ]"

import xlrd
import sys
import argparse
import time

class FileFormatException(Exception):
    def __init__(self, value):
        self.value = value

class Data_setup:
    hosts_d = {}        # Dictionary which stores host lists.
    hosts_l = []        # Host list.
    host_vars = {}      # Dictionary which stores attributes of each host.
	
    def __init__(self):
    # Read data fron the Excel file.
    # The Excel file is opened with read only.
        try:
            book = xlrd.open_workbook(args.infile)      # Create Book from the Excel
        except Exception as e:
            print "Open_workbook error: " + args.infile
            raise e
#       print book.nsheets
#       print book.sheet_by_index(0).name
        first_sheet = book.sheet_by_index(0).name
        self.sheet = book.sheet_by_name(first_sheet)  # Get the first sheet.

    def data_setup(self):

        line = 1
        for row in range(1, self.sheet.nrows): 
            line += 1
            hostname = self.sheet.cell(row,0).value  
            hostname = hostname.strip()
            if not hostname: 
                raise FileFormatException(line)

            # Check required cell
            ip = self.sheet.cell(row,2).value
            ip = ip.strip()
            if not ip:
                if args.type == "cobbler":
                    raise FileFormatException(line)

            bmc = self.sheet.cell(row,3).value
            bmc = bmc.strip()
            if not bmc:
                if args.type == "power_bmc":
                    raise FileFormatException(line)

            mac = self.sheet.cell(row,4).value
            mac = mac.strip()
            if not mac:
                if args.type == "cobbler" or args.type == "power_wol":
                    raise FileFormatException(line)

            mask = self.sheet.cell(row,5).value
            mask = mask.strip()
            if not mask:
                if args.type == "cobbler":
                    raise FileFormatException(line)


            gate = self.sheet.cell(row,6).value
            gate = gate.strip()

            # Store data into host table.
            self.hosts_l.append(hostname)  
            self.host_vars[hostname] = {  
                'ip_addr': ip,
                'bmc_ip_addr': bmc,
                'mac_addr': mac,
                'netmask': mask,
                'gateway': gate,
            }

class  ansible_method(Data_setup):

    def data_setup(self):
        line = 1
        for row in range(1, self.sheet.nrows): 	# skip title line
            line += 1
            hostname = self.sheet.cell(row,0).value  
            hostname = hostname.strip()
            if not hostname: 
                raise FileFormatException(line)

            if self.sheet.cell(row,1).value:
                group = self.sheet.cell(row,1).value	
            else:
                group = ""
            if group not in self.hosts_d:
                self.hosts_d[group] = {"hosts": []}

            self.hosts_d[group]["hosts"].append(hostname)  

    def data_output(self):
    # Output sorted hosts_d
    # The type of the return value of ssorted() is double.
        for g, h in sorted(self.hosts_d.items()):
            if g != "":
                args.outfile.write('['+ g +']\n')
            for j in h['hosts']:
                args.outfile.write(j + '\n')
 

class cobbler_method(Data_setup):

	def data_output(self):
		for i in self.hosts_l:
			args.outfile.write(i + ' ' + self.host_vars[i]['mac_addr'] + ' ' 
					+ self.host_vars[i]['ip_addr'] + ' ' + self.host_vars[i]['netmask'] + ' ' 
                        + self.host_vars[i]['gateway'] + '\n')

 
class powerctrl_method(Data_setup):

	def data_output(self):
		for i in self.hosts_l:
			args.outfile.write(i + ' ' 
					+ self.host_vars[i]['bmc_ip_addr'] + '\n')

class powerctrl_w_method(Data_setup):

	def data_output(self):
		for i in self.hosts_l:
			args.outfile.write(i + ' ' 
				+ self.host_vars[i]['mac_addr'] + '\n')

def create_hostlist():

    try:
        if args.type == "ansible":
            setup = ansible_method()
        elif args.type=="cobbler":
            setup = cobbler_method()
        elif args.type == "power_bmc":
            setup = powerctrl_method()
        elif args.type == "power_wol":
            setup = powerctrl_w_method()
    except IOError as ie:
        print "I/O error ({0}): {1}".format(ie.errno, ie.strerror)
        raise ie
    except Exception as e:
        print "Error ({0}): {1}".format(ie.errno, ie.strerror)
        raise e

    # Data input from Excel
    try:
        setup.data_setup();
    except FileFormatException as fe:
        print "File format error in line " + str(fe.value) + "."
        sys.exit(1)
    # Data output
    setup.data_output()



if __name__ == '__main__':
    import argparse
    import os 
    XLS_DEFAULT = '/opt/nec/sysmng-soft/etc/host-list.xlsx'

    p = argparse.ArgumentParser()
    p.add_argument("--type", required=True, # Required option
        choices=['ansible', 'cobbler', 'power_bmc', 'power_wol'],
        help='type of the destination file')	
    p.add_argument("--infile", default=XLS_DEFAULT, 
        help='input filename (default: ' + XLS_DEFAULT + ')')	
    p.add_argument("--outfile", type=argparse.FileType('w+x'),
        default=sys.stdout, help='output filename (default: stdout)')	

    args = p.parse_args()

    if not os.path.exists(args.infile):
        print args.infile + " not found." 
        sys.exit(1)

    try:
        create_hostlist()
    except Exception as e:
        sys.exit(1)
