###
###  VE H/W metrics monitoring module
###  Copyright (c) 2018 NEC Corporation
###

import subprocess
import traceback, sys
import os.path
import glob
import copy
import threading
import string
import os
import re 
import time 
import ve_common
import ymlparse
import logging
import logging.handlers
from logging import Formatter

f_value = {}
descriptors = []
sysfs_file_dir = {}
ve_logger = None
last_update = 0
_Worker_Thread = None
_Lock = threading.Lock() # synchronization lock

# Metric definition
METRICS_F = [
    { "name": 've_num',  "time_max": 60,  "value_type": 'uint',  "units": 'VEs', 
      "slope": 'zero',   "format" : "%d", "description": 'Attached_VEs'     },
    { "name": 'mmm_ver', "time_max": 60,  "value_type": 'string', "units": '', 
      "slope": 'zero',   "format" : "%s", "description": 'MMM Version'       },
    ]

VEMETRICS_F = [
    { "name": 'model',  "time_max": 60,  "value_type": 'uint', "units": '',
      "slope": 'zero',   "format" : "%d", "description": 'VE Model' },
    { "name": 'product_type',  "time_max": 60,  "value_type": 'uint', "units": '',
      "slope": 'zero',   "format" : "%d", "description": 'Product Type' },
    { "name": 'cores',  "time_max": 60,  "value_type": 'uint', "units": 'Cores', 
      "slope": 'zero',   "format" : "%d", "description": 'Cores'   },
    { "name": 'chip_serial', "time_max": 60,  "value_type": 'string', "units": '',
      "slope": 'zero',   "format" : "%s", "description": 'VE Chip S/N'},
    { "name": 'board_serial', "time_max": 60,  "value_type": 'string', "units": '',
      "slope": 'zero',   "format" : "%s", "description": 'VE Board S/N'},
    { "name": 'driver_ver', "time_max": 60,  "value_type": 'string', "units": '',
      "slope": 'zero',   "format" : "%s", "description": 'VE Driver Version'},
    { "name": 'vmcfw_ver', "time_max": 60,  "value_type": 'string', "units": '',
      "slope": 'zero',   "format" : "%s", "description": 'VMCFW Version'},
    { "name": 'core_clock', "time_max": 60,  "value_type": 'uint', "units": 'MHz', 
      "slope": 'zero',   "format" : "%d", "description": 'Core Clock'   },
    { "name": 'base_clock', "time_max": 60,  "value_type": 'uint', "units": 'MHz', 
      "slope": 'zero',   "format" : "%d", "description": 'Base Clock'   },
    { "name": 'memory_clock', "time_max": 60,  "value_type": 'uint', "units": 'MHz', 
      "slope": 'zero',   "format" : "%d", "description": 'Memory Clock'   },
    { "name": 'memory_size', "time_max": 60,  "value_type": 'uint', "units": 'GB', 
      "slope": 'zero',   "format" : "%d", "description": 'Memory Size'   },
    { "name": 'l1_icache', "time_max": 60,  "value_type": 'uint', "units": 'KB/core', 
      "slope": 'zero',   "format" : "%d", "description": 'L1 I-Cache Size'   },
    { "name": 'l1_dcache', "time_max": 60,  "value_type": 'uint', "units": 'KB/core', 
      "slope": 'zero',   "format" : "%d", "description": 'L1 D-Cache Size'   },
    { "name": 'l2_cache', "time_max": 60,  "value_type": 'uint', "units": 'KB/core', 
      "slope": 'zero',   "format" : "%d", "description": 'L2 Cache Size'   },
    { "name": 'llc_cache', "time_max": 60,  "value_type": 'uint', "units": 'MB/VE', 
      "slope": 'zero',   "format" : "%d", "description": 'LLC Cache Size'   },
    { "name": 'partition_mode', "time_max": 60,  "value_type": 'string', "units": '',
      "slope": 'zero',   "format" : "%s", "description": 'Partitioning Mode'},
    { "name": 'throttle_level', "time_max": 60,  "value_type": 'uint', "units": '',
      "slope": 'zero',   "format" : "%d", "description": 'Throttling Level'},
    { "name": 'pci_bus', "time_max": 60,  "value_type": 'string', "units": '',
      "slope": 'zero',   "format" : "%s", "description": 'PCI Bus ID'},
    { "name": 'pci_device', "time_max": 60,  "value_type": 'string', "units": '',
      "slope": 'zero',   "format" : "%s", "description": 'PCI Device ID'},
    ]

VEMETRICS = [
    { "name": 've_state', "time_max": 60,  "value_type": 'string', "units": '', 
      "slope": 'both',   "format" : "%s", "description": 'VE State' },
    { "name": 've_core0_temp',  "time_max": 60,  "value_type": 'float', "units": 'C', 
      "slope": 'both',   "format" : "%.1f", "description": 'Temperature Chip Core0'  },
    { "name": 've_temp_ve_diode_0',  "time_max": 60,  "value_type": 'float', "units": 'C', 
      "slope": 'both',   "format" : "%.1f", "description": 'Temperature Chip Diode 0'  },
    { "name": 've_temp_ve_diode_1',  "time_max": 60,  "value_type": 'float', "units": 'C', 
      "slope": 'both',   "format" : "%.1f", "description": 'Temperature Chip Diode 1'  },
#    { "name": 've_fan',"time_max": 60,  "value_type": 'uint', "units": 'RPM', 
#      "slope": 'both',   "format" : "%d", "description": 'Fan Speed'  },
    { "name": 've_core_vdd0',"time_max": 60,  "value_type": 'float', "units": 'mV', 
      "slope": 'both',   "format" : "%.1f", "description": 'Voltage Core VDD 0'  },
    { "name": 've_core_vdd1',"time_max": 60,  "value_type": 'float', "units": 'mV', 
      "slope": 'both',   "format" : "%.1f", "description": 'Voltage Core VDD 1'  },
    ]

DEFAULT_REFRESH = 60
DEFAULT_YML = "/etc/opt/nec/ve/mmm/info/ve_hw_spec.yaml"
NAME_PREFIX = 've_'

#logging message
FILE_OPEN_FAILED     = " Can not open "
FILE_READ_FAILED     = " Can not read "
FILE_CLOSE_FAILED    = " Can not close "
THREAD_ERROR_JOIN    = " Can not shutdown thread."
THREAD_ERROR_RELEASE = " Can not release thread."
THREAD_ERROR_START   = " Can not start thread."

class UpdateMetricThread(threading.Thread):

    def __init__(self, params, f_metric):
        threading.Thread.__init__(self)
        self.running      = False
        self.shuttingdown = False
        self.refresh_rate = DEFAULT_REFRESH
        if "RefreshRate" in params:
            self.refresh_rate = int(params["RefreshRate"])
        self.yamlFile = DEFAULT_YML
        if "SensorTable" in params:
            self.yamlFile = params["SensorTable"]
        self.metric       = f_metric
        ymlparse.readYaml(self.yamlFile)

    def shutdown(self):
        self.shuttingdown = True
        if not self.running:
            return
        try:
            self.join()
        except Exception as e:
            output_logmsg(e, THREAD_ERROR_JOIN)
            os._exit(1)

    def run(self):
        self.running = True
        while not self.shuttingdown:
            self.update_metric()
            time.sleep(self.refresh_rate)
            
        self.running = False

    def update_metric(self):
        ve_list = self.metric['ve_list']
        for desc in VEMETRICS:
            for ve_n in ve_list:
                sysfs_dir = sysfs_file_dir['ve_'+str(ve_n)+'sysfsd']
                metric_name = desc['name']
                
                isVeStateOnline = self.get_ve_state(sysfs_dir, ve_n)
                
                if metric_name != 've_state':
                    #VE state is ONLINE
                    if isVeStateOnline == True:
                        self.get_value(ve_n, sysfs_dir, metric_name, desc)
                        
                    #VE state is OFFLINE
                    else:
                        self.metric[NAME_PREFIX +  str(ve_n) + "_" + metric_name] = 0
    
    def get_ve_state(self, sysfs, ve_n):
        file_name = "ve_state"
        metric_name = NAME_PREFIX +  str(ve_n) + "_" + "ve_state"
        isOnline = False
        
        try:
           file = open(sysfs + "/" + file_name, "r")
        except Exception as e:
            output_logmsg(e, FILE_OPEN_FAILED + "ve_state" + " metric sensor file : " + sysfs + "/" + file_name)
            self.metric[metric_name] = 'UNKNOWN'
            return False
        else:
            try:
                line = file.readline()
            except Exception as e:
                output_logmsg(e, FILE_READ_FAILED + "ve_state" + " metric sensor file : " + sysfs + "/" + file_name)
                self.metric[metric_name] = 'UNKNOWN'
                return False
            else:
	        val = re.split("\s+", line)
	        _Lock.acquire()
	        
	        if val[0] == '0':
	            self.metric[metric_name] = 'UNINITIALIZED'
	        elif val[0] == '1':
	            self.metric[metric_name] = 'ONLINE'
	            isOnline = True
	        elif val[0] == '2':
	            self.metric[metric_name] = 'OFFLINE'
	        elif val[0] == '3':
	            self.metric[metric_name] = 'MAINTENANCE'
	        elif val[0] == '4':
	            self.metric[metric_name] = 'UNAVAILABLE'
	        else:
	            self.metric[metric_name] = 'UNKNOWN'
                    
	        try: 
	            _Lock.release()
	        except Exception as e:
	            output_logmsg(e, THREAD_ERROR_RELEASE)
	            self.metric[metric_name] = 'UNKNOWN'
	            return False
	        finally:
	            try:
	                file.close()
	            except Exception as e:
	                output_logmsg(e, FILE_CLOSE_FAILED + "ve_state" + " metric sensor file : " + sysfs + "/" + file_name)
	                self.metric[metric_name] = 'UNKNOWN'
	                return False
	            else:
	                return isOnline
	                    
    def get_value(self, ve_n, sysfs, m_name, desc):
        metric_name=NAME_PREFIX +  str(ve_n) + "_" + m_name

        if m_name.startswith("ve_core") and m_name.endswith("_temp"):
            lnum = int(m_name[7])
            pnum = physical_corenum(ve_n, lnum)
            
            #Logical to physical conversion of CPU number failed
            if pnum < 0:
                ve_logger.warning("Logical / physical conversion of CPU number failed")
                self.metric[metric_name] = 0
                return
            s_name = "ve_core" + str(pnum) + "_temp"
        else:
            s_name = m_name
            
        model_key = NAME_PREFIX + str(ve_n) + '_' + 'model'
        type_key  = NAME_PREFIX + str(ve_n) + '_' + 'product_type'
        
        ve_type = 've' + str(self.metric[model_key]) + '_' + str(self.metric[type_key])
        
        p = ymlparse.getSysFileName(s_name, ve_type)
        if p != None:
            file_name = p
        else:
            self.metric[metric_name] = 0
            ve_logger.warning("Can not parse "+m_name+" yamlfile.")
            return
        
        try:
            file = open(sysfs + "/" + file_name, 'r')
        except Exception as e:
            self.metric[metric_name] = 0
            output_logmsg(e, FILE_OPEN_FAILED + m_name + " metric sensor file : " + sysfs + "/" + file_name)
            return
        else:
            try:
                line = file.readline()
            except Exception as e:
                self.metric[metric_name] = 0
                output_logmsg(e, FILE_READ_FAILED + m_name + " metric sensor file : " + sysfs + "/" + file_name)
                return
            else:
                val = re.split("\s+", line)
	        
	        _Lock.acquire()
	        form = desc['format']
	        if  form == '%d' :
	            self.metric[metric_name] = int(val[0])
	        elif form == '%.1f':
	            unit = desc['units']
	            if unit == 'C':
	                self.metric[metric_name] = float(val[0]) / 1000000.0
	            else: # Voltage
	                self.metric[metric_name] = float(val[0]) / 1000.0
	        else:
                    self.metric[metric_name] = val[0]
                try: 
	            _Lock.release()
	        except Exception as e:
                    self.metric[metric_name] = 0
    	            output_logmsg(e, THREAD_ERROR_RELEASE)                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  
	            return
	        finally:
	            try:
	                file.close()
	            except Exception as e:
	                self.metric[metric_name] = 0
	                output_logmsg(e, FILE_CLOSE_FAILED + m_name + " metric sensor file : " + sysfs + "/" + file_name)
	                return

    def metric_of(self, name):
        val = 0
        if name in self.metric:
            _Lock.acquire()
            if self.metric[name] != None:
                val = self.metric[name]
                ve_logger.debug(name + " = " + str(self.metric[name]))
            try: 
                _Lock.release()
            except Exception as e:
                output_logmsg(e, THREAD_ERROR_RELEASE)
        return val
    
def physical_corenum(ve, cnum):
    sysfs_dir = sysfs_file_dir['ve_'+str(ve)+'sysfsd']
    sysfs_file = sysfs_dir + "/cores_enable"
    pnum = -1
    found = 0
    
    try:
        file = open(sysfs_file, 'r')
    except Exception as e:
        output_logmsg(e, FILE_OPEN_FAILED + sysfs_file)
        return pnum
    else:
        try:
            line = file.readline()
        except Exception as e: 
            output_logmsg(e, FILE_READ_FAILED + sysfs_file)
            return pnum
        else:
	        val = re.split("\s+", line)
	        cores = int(val[0], 16)
	        lnum = cnum
	        
	        while cores != 0:
	            if (cores & 1) == 1:
	                if lnum == 0:
	                    found = 1
	                    break
	                else:
	                    lnum = lnum - 1
	            pnum = pnum + 1
	            cores = cores >> 1

	        if found == 0:
	            pnum = -1
	        else:
	            pnum += 1
	        
	        try:
	            file.close()
	        except Exception as e:
	            output_logmsg(e, FILE_CLOSE_FAILED + sysfs_file)
	            return pnum
	        else:
	            return pnum

def output_logmsg(e, msg):
    #if exception attribute has errno, output errno
    if hasattr(e, "errno") == True: 
        ve_logger.warning("errno=" + str(e.errno) + msg)
    else:
        ve_logger.warning(msg)

def metric_of(name):
    return _Worker_Thread.metric_of(name)

def init_veinfo(metric, ve_list):
    global sysfs_file_dir
    SYSTEM_DIR = "/sys"
    
    f_value = ve_common.get_from_vecmd(metric)
    if f_value == {}:
        ve_logger.warning("get from vecmd failed.")
        return {}
    
    for ve in ve_list:
        dir = ve_common.get_sysdir(ve).rstrip("\n")
        
        if dir == "":
            ve_logger.warning("Executing udevadm command failed.")
            return {}
        elif dir[0] != '/':
            ve_logger.warning("Executing udevadm command failed.")
            return {}
            
        sys_dir = SYSTEM_DIR + dir
        sysfs_file_dir['ve_'+str(ve)+'sysfsd'] = sys_dir
        
    return f_value

def metric_init(params):
    global descriptors, _Worker_Thread, ve_logger

    ve_logger = logging.getLogger('VE_LOGGER')
    ve_logger.setLevel(logging.DEBUG)
    
    handler = logging.handlers.SysLogHandler(address = "/dev/log")
    formatter = Formatter(fmt='%(levelname)s - VEHW : %(message)s')
    handler.formatter = formatter
    ve_logger.addHandler(handler)
    
    f_metric = {}
    f_metric.update(ve_common.get_ve_base_info())
    if len(f_metric) == 0:
        ve_logger.warning("Getting the number of VE failed.")
        return {}
    ve_list = f_metric['ve_list']
    
    vecmd_metric = init_veinfo(VEMETRICS_F, ve_list)
    if len(vecmd_metric) == 0:
        ve_logger.warning("Getting the vecmd metrics failed.")
        return {}
    f_metric.update(vecmd_metric)
    
    for metric in METRICS_F:
        metric['call_back'] = metric_of
        descriptors.append(metric)
    
    for ve in ve_list:
        for metric in VEMETRICS_F:
            d = copy.copy(metric)
            d['name'] = NAME_PREFIX + str(ve) + '_' + d['name']
            d['call_back'] = metric_of
            d['groups'] = "ve" +  str(ve)
            descriptors.append(d)
        for metric in VEMETRICS:
            d = copy.copy(metric)
            d['name'] = NAME_PREFIX + str(ve) + '_' + d['name']
            d['call_back'] = metric_of
            d['groups'] = "ve" +  str(ve)
            descriptors.append(d)

    _Worker_Thread = UpdateMetricThread(params, f_metric)
    try:
        _Worker_Thread.start()
    except Exception as e:
        output_logmsg(e, THREAD_ERROR_START)
        os._exit(1)
    else:
        return descriptors 

def metric_cleanup():
    _Worker_Thread.shutdown()

#This code is for debugging and unit testing
if __name__ == '__main__':
    try:
        metric_init({})
        
        f_metric = (ve_common.get_ve_base_info())
        if len(f_metric) == 0:
            os._exit(1)
        ve_list = f_metric['ve_list']
        
        for ve in ve_list:
            while True:
                for d in descriptors:
                    v = d['call_back'](d['name'])
                    print ('value for %s is '+d['format']) % (d['name'],  v)
                time.sleep(5)
                
    except KeyboardInterrupt:
        time.sleep(0.2)
        os._exit(1)
    except:
        traceback.print_exc()
        os._exit(1)
