import subprocess
import re

NAME_PREFIX='ve_'

def get_ve_base_info():
    ve_list = []
    val = {}
    command =  "/opt/nec/ve/bin/vecmd info"
    proc = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    pout, perr = proc.communicate()
    ecode = proc.wait()
    
    if ecode != 0:
        return val
    
    for m in re.split('(?:\r\n|\n)',pout):
        d = m.split()
        if len(d) == 4 and d[0] == "Attached" and d[1] == "VEs":
            val['ve_num'] = int(d[3])
            continue
        
        elif len(d) == 4 and d[0] == "MMM" and d[1] == "Version":
            val['mmm_ver'] = d[3]
            continue
        
        elif len(d) == 4 and d[0][0] == "[" and d[0][1] == "V" and d[0][2] == "E" and d[1] == ":":
            ve_str = d[0].split("VE")
            ve_list.append(int(ve_str[1]))
            continue
    
    val['ve_list'] = ve_list
    
    if val['ve_num'] != len(val['ve_list']):
        return {}
    
    return val

def get_from_vecmd(metric):
    val = {}
    command =  "/opt/nec/ve/bin/vecmd info"

    proc = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    pout, perr = proc.communicate()
    ecode = proc.wait()
    
    if ecode != 0:
        return val
        
    for m in re.split('(?:\r\n|\n)',pout):
        d = m.split()
        
        if len(d) == 4 and d[0][0] == "[" and d[0][1] == "V" and d[0][2] == "E" and d[1] == ":":
            ve_str = d[0].split("VE")
            ve_no = int(ve_str[1])
            continue

        for m in metric:
            if 'model' == m['name']:
                if len(d) ==  4 and d[0] == "VE" and d[1] == "Model":
                    val[NAME_PREFIX + str(ve_no) + '_' + m['name']] = int(d[3])
                    continue

            if 'product_type' == m['name']:
                if len(d) ==  4 and d[0] == "Product" and d[1] == "Type":
                    val[NAME_PREFIX + str(ve_no) + '_' + m['name']] = int(d[3])
                    continue

            if 'cores' == m['name']:
                if len(d) ==  3 and d[0] == "Cores" and d[1] == ":":
                    val[NAME_PREFIX + str(ve_no) + '_' + m['name']] = int(d[2])
                    
                    #If VE cores is not 10, partition mode is not displayed by vecmd.
                    if int(d[2]) != 10:
                        val[NAME_PREFIX + str(ve_no) + '_' + 'partition_mode'] = "UNAVAILABLE"                    
                    continue
        
            if 'chip_serial' == m['name']:
                if len(d) ==  5 and d[0] == "VE" and d[1] == "Chip" and d[2] == "S/N":
                    val[NAME_PREFIX + str(ve_no) + '_' + m['name']] = d[4]
                    continue

            if 'board_serial' == m['name']:
                if len(d) ==  5 and d[0] == "VE" and d[1] == "Board" and d[2] == "S/N":
                    val[NAME_PREFIX + str(ve_no) + '_' +m['name']] = d[4]
                    continue

            if 'driver_ver' == m['name']:
                if len(d) ==  5 and d[0] == "VE" and d[1] == "Driver" and d[2] == "Version":
                    val[NAME_PREFIX + str(ve_no) + '_' +m['name']] = d[4]
                    continue

            if 'vmcfw_ver' == m['name']:
                if len(d) ==  4 and d[0] == "VMCFW" and d[1] == "Version" and d[2] == ":":
                    val[NAME_PREFIX + str(ve_no) + '_' +m['name']] = d[3]
                    continue

            if 'core_clock' == m['name']:
                if len(d) ==  5 and d[0] == "Core" and d[1] == "Clock" and d[2] == ":":
                    val[NAME_PREFIX + str(ve_no) + '_' +m['name']] = int(d[3])
                    continue

            if 'base_clock' == m['name']:
                if len(d) ==  5 and d[0] == "Base" and d[1] == "Clock" and d[2] == ":":
                    val[NAME_PREFIX + str(ve_no) + '_' +m['name']] = int(d[3])
                    continue

            if 'memory_clock' == m['name']:
                if len(d) ==  5 and d[0] == "Memory" and d[1] == "Clock" and d[2] == ":":
                    val[NAME_PREFIX + str(ve_no) + '_' +m['name']] = int(d[3])
                    continue

            if 'memory_size' == m['name']:
                if len(d) ==  5 and d[0] == "Memory" and d[1] == "Size" and d[2] == ":":
                    val[NAME_PREFIX + str(ve_no) + '_' +m['name']] = int(d[3])
                    continue

            if 'l1_icache' == m['name']:
                if len(d) ==  6 and d[0] == "L1" and d[1] == "I-Cache" and d[2] == "Size":
                    val[NAME_PREFIX + str(ve_no) + '_' +m['name']] = int(d[4])
                    continue

            if 'l1_dcache' == m['name']:
                if len(d) ==  6 and d[0] == "L1" and d[1] == "D-Cache" and d[2] == "Size":
                    val[NAME_PREFIX + str(ve_no) + '_' +m['name']] = int(d[4])
                    continue

            if 'l2_cache' == m['name']:
                if len(d) ==  6 and d[0] == "L2" and d[1] == "Cache" and d[2] == "Size":
                    val[NAME_PREFIX + str(ve_no) + '_' +m['name']] = int(d[4])
                    continue

            if 'llc_cache' == m['name']:
                if len(d) ==  6 and d[0] == "LLC" and d[1] == "Cache" and d[2] == "Size":
                    val[NAME_PREFIX + str(ve_no) + '_' +m['name']] = int(d[4])
                    continue

            if 'partition_mode' == m['name']:
                if len(d) ==  4 and d[0] == "Partitioning" and d[1] == "Mode" and d[2] == ":":
                    val[NAME_PREFIX + str(ve_no) + '_' +m['name']] = d[3]
                    continue

            if 'throttle_level' == m['name']:
                if len(d) ==  4 and d[0] == "Throttling" and d[1] == "Level" and d[2] == ":":
                    val[NAME_PREFIX + str(ve_no) + '_' +m['name']] = int(d[3])
                    continue

            if 'pci_bus' == m['name']:
                if len(d) ==  4 and d[0] == "Bus" and d[1] == "ID" and d[2] == ":":
                    val[NAME_PREFIX + str(ve_no) + '_' +m['name']] = d[3]
                    continue

            if 'pci_device' == m['name']:
                if len(d) ==  4 and d[0] == "Device" and d[1] == "ID" and d[2] == ":":
                    val[NAME_PREFIX + str(ve_no) + '_' +m['name']] = d[3]
                    continue
    return val
        
def get_sysdir(ve_n):
    command = "/sbin/udevadm info --query=path --name=veslot" + str(ve_n)
    
    proc = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    sysfs_ret, sysfs_eret = proc.communicate()
    ecode = proc.wait()
    if ecode == 0 :
        return sysfs_ret
    else:
        return ""
